#!/usr/bin/env python
"""
This example demonstrates how to use the FUN3D python interfaces to
set modal displacements read modal states on the fly.
"""
import f90nml
from mpi4py import MPI
from fun3d.solvers import Flow
from fun3d import interface


def get_modal_displacement(step, nsteps):
    return 0.01 * (step / nsteps) + 0.01


def main():
    command_line_options = {"aeroelastic_internal": True}

    nml = f90nml.read("fun3d.nml")
    nsteps = nml["code_run_control"]["steps"]

    # ==========================================================================
    # Initialize the fun3d flow solver - read namelist, partion mesh, etc.
    # ==========================================================================
    comm = MPI.COMM_WORLD

    flow = Flow()
    flow.initialize(kwargs=command_line_options, comm=comm)

    # ==========================================================================
    # Time step loop
    # ==========================================================================
    body = 1
    for step in range(1,nsteps+1):

        # prescribe the displacement of mode 1 (1-based numbering for FUN3D modes)
        gdisp = get_modal_displacement(step, nsteps)
        mode = 1
        if comm.rank == 0:
            print("gdisp", gdisp)
        interface.aeroelastic_push_modal_displacement(body, mode, gdisp)

        # take a flow step
        flow.iterate()

        # read the modal state of mode 2
        mode = 2
        state = interface.aeroelastic_pull_modal_state(body, mode)

        if comm.Get_rank() == 0:
            print(
                f"Modal state at step {step}\n",
                f"   gdisp  = {state[0]}\n",
                f"   gvel   = {state[1]}\n",
                f"   gaccel = {state[2]}\n",
                f"   gforce = {state[3]}\n",
            )

    # ==========================================================================
    # Finish the fun3d analysis. Write restart files, hist files, etc
    # ==========================================================================
    flow.post()


if __name__ == "__main__":
    main()
