#!/usr/bin/env python
import f90nml
import numpy as np
from mpi4py import MPI
from fun3d.solvers import Flow


def get_rigid_motion(step):
    """
    rigid motion is sinusoidal pitching and plunging motion
    """
    # standard fun3d transform 4x4 for rigid motion
    transform = np.eye(4,order="F")

    # rotation piece
    theta = 3.0 * np.pi / 180.0 * np.sin(2.0 * np.pi * float(step) / 50.0)
    transform[0, 0] = np.cos(theta)
    transform[0, 2] = -np.sin(theta)
    transform[2, 0] = np.sin(theta)
    transform[2, 2] = np.cos(theta)

    # translation piece
    transform[2, -1] = np.sin(2.0 * np.pi * float(step / 50.0))
    return transform


def get_deform_motion(step, x, y, z):
    """
    deforming motion is parabolic in space, sinusoidal in time
    """

    dx = np.zeros_like(x)
    dy = np.zeros_like(y)
    dz = np.zeros_like(z) + 1e-6

    dz = 0.1 * y**2 * np.sin(2.0 * np.pi * float(step) / 50.0)
    return dx, dy, dz


def get_wall_temperature(step, x, y, z):
    """
    wall temperature is parabolic in space, sinusoidal in time
    """
    temperature = 1.2 + 0.1 * y**2 * np.sin(2.0 * np.pi * float(step) / 50.0)
    return temperature


def main():
    # Demo script options
    prior_steps = 0  # if restarting set to number of steps in the restart file
    nml = f90nml.read("fun3d.nml")
    nsteps = nml["code_run_control"]["steps"]

    moving_body_input = f90nml.read("moving_body.input")
    mesh_movement = moving_body_input["body_definitions"]["mesh_movement"][0]
    thermal_coupling = True
    command_line_options = {}

    # ==========================================================================
    # Initialize the fun3d flow solver - read namelist, partion mesh, etc.
    # ==========================================================================
    comm = MPI.COMM_WORLD

    flow = Flow()
    flow.initialize(kwargs=command_line_options, comm=comm)

    # ==========================================================================
    # Pull out the surface node information for coupling
    # ==========================================================================
    # body number for multibody problems. this demo only has 1 body.
    ibody = 1  # fun3d numbering so starts at 1

    # Each processor determines how many surface nodes are owned by that proc.
    # Not all processors may have surface nodes.
    # All vectors in and out of fun3d are local-owned values only.
    # MPI operations like gatherv and scatterv can be used to create/distribute a global image of the coordinates/displacements/forces
    num_nodes = flow.extract_surface_num(body=ibody)

    # get the initial coordinates
    x = np.zeros(num_nodes)
    y = np.zeros(num_nodes)
    z = np.zeros(num_nodes)
    if num_nodes > 0:
        x, y, z = flow.extract_surface(num_nodes, body=ibody)

    # ==========================================================================
    # Time step loop
    # ==========================================================================
    for step in range(prior_steps + 1, prior_steps + 1 + nsteps):

        # input the motion
        if "rigid" in mesh_movement:
            transform = get_rigid_motion(step)

            flow.input_rigid_transform(transform, body=ibody)

        if num_nodes > 0:
            if "deform" in mesh_movement:
                dx, dy, dz = get_deform_motion(step, x, y, z)
                flow.input_deformation(dx, dy, dz, body=ibody)
            if thermal_coupling:
                temperature = get_wall_temperature(step, x, y, z)
                # input wall temperature should be normalized: T/T_ref
                flow.input_wall_temperature(temperature, body=ibody)

        # take a step of fun3d
        flow.iterate()

        # pull out the surface forces. These are nondimensional and at the nodes.
        # multiply by dynamic pressure to get dimensional forces.
        # Forces in inertial frame by default only includes pressure forces.
        if num_nodes > 0:
            fx, fy, fz = flow.extract_forces(num_nodes, body=ibody)
        else:
            fx = 0.0
            fy = 0.0
            fz = 0.0

        fx_total = comm.allreduce(np.sum(fx))
        fy_total = comm.allreduce(np.sum(fy))
        fz_total = comm.allreduce(np.sum(fz))

        if comm.Get_rank() == 0:
            print("Total surface force", fx_total, fy_total, fz_total)

        # pull out the heat fluxes*area. These are nondimensional and at the nodes.
        # multiply by freesteam reference power per area (1/2*rho_ref*U_ref^3) to get dimensional heat flux*area.
        # heat flux only for viscous surfaces in the compressible path
        if num_nodes > 0:
            cqx, cqy, cqz = flow.extract_heat_flux(num_nodes, body=ibody)
        else:
            cqx = 0.0
            cqy = 0.0
            cqz = 0.0

        cqx_total = comm.allreduce(np.sum(cqx))
        cqy_total = comm.allreduce(np.sum(cqy))
        cqz_total = comm.allreduce(np.sum(cqz))

        if comm.Get_rank() == 0:
            print("Total surface heat flux*area", cqx_total, cqy_total, cqz_total)

    # ==========================================================================
    # Finish the fun3d analysis. Write restart files, hist files, etc
    # ==========================================================================
    flow.post()


if __name__ == "__main__":
    main()
